﻿using UnityEngine;
using System.Collections.Generic;
using SimpleJSON;

namespace MacGruber
{
    public class RandomBlend : MVRScript
    {
		private JSONStorableBool myAcceptTrigger;
		private JSONStorableFloat myMinTime;
		private JSONStorableFloat myMaxTime;
		private JSONStorableBool myBlendInOut;
		private JSONStorableBool myBlendSmooth;
		private EventTrigger myStartTrigger;
		private FloatTrigger myValueTrigger;
		private EventTrigger myEndTrigger;

		private float myClock = -1.0f;
		private float myDuration = 0.0f;

        public override void Init()
		{
			Utils.SetupInfoText(this,
				"<color=#606060><size=40><b>RandomBlend</b></size>\nLogicBrick similar to a <i>RandomDelay</i> brick. However, after <i>Trigger</i> was called, it provides separate <i>Start</i> trigger, a value slowly going from 0 to 1 and eventually an <i>End</i> trigger. Classic use case would be to blend something in or out.\n\n" +
				"<b>Blend In+Out:</b> Instead of transitioning from 0 to 1, value reaches 1 at the midpoint and then goes back down to 0.\n\n" +
				"<b>Blend Smooth:</b> Apply classic SmoothStep curve to the value.</color>\n\n",
				600.0f, true
			);
			
			myAcceptTrigger = Utils.SetupToggle(this, "AcceptTrigger", true, false);

			myMinTime = Utils.SetupSliderFloatWithRange(this, "Min Time (sec)", 3, 0, 10, false);
			myMinTime.constrained = false;
			myMinTime.setCallbackFunction += (float v) => {
				Utils.AdjustMaxSliderFromMin(v, myMaxTime);
			};

			myMaxTime = Utils.SetupSliderFloatWithRange(this, "Max Time (sec)", 5, 0, 10, false);
			myMaxTime.constrained = false;
			myMaxTime.setCallbackFunction += (float v) => {
				myMinTime.valNoCallback = Mathf.Min(v, myMinTime.val);
			};

			myBlendInOut = Utils.SetupToggle(this, "Blend In+Out", false, false);
			myBlendSmooth = Utils.SetupToggle(this, "Blend Smooth", true, false);

			myStartTrigger = new EventTrigger(this, "OnStartTrigger");
			myValueTrigger = new FloatTrigger(this, "OnValueTrigger");
			myEndTrigger = new EventTrigger(this, "OnEndTrigger");
			Utils.SetupButton(this, "Actions OnStartTrigger", myStartTrigger.OpenPanel, false);
			Utils.SetupButton(this, "Actions OnValueTrigger", myValueTrigger.OpenPanel, false);
			Utils.SetupButton(this, "Actions OnEndTrigger", myEndTrigger.OpenPanel, false);

			Utils.SetupButton(this, "Test Trigger", Trigger, false);
			Utils.SetupAction(this, "Trigger", Trigger);

			SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
			SimpleTriggerHandler.LoadAssets();
		}

		private void OnAtomRename(string oldid, string newid)
		{
			myStartTrigger.SyncAtomNames();
			myValueTrigger.SyncAtomNames();
			myEndTrigger.SyncAtomNames();
		}

		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if (includePhysical || forceStore)
			{
				needsStore = true;
				jc[myStartTrigger.Name] = myStartTrigger.GetJSON(base.subScenePrefix);
				jc[myValueTrigger.Name] = myValueTrigger.GetJSON(base.subScenePrefix);
				jc[myEndTrigger.Name] = myEndTrigger.GetJSON(base.subScenePrefix);
			}
			return jc;
		}

		public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
		{
			base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
			Utils.AdjustSliderRange(myMinTime);
			Utils.AdjustSliderRange(myMaxTime);
			myClock = -1.0f;

			if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
			{
				myStartTrigger.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
				myValueTrigger.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
				myEndTrigger.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
			}
		}

		private void OnDestroy()
		{
			SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;

			myStartTrigger.Remove();
			myValueTrigger.Remove();
			myEndTrigger.Remove();
		}

		private void Update()
		{
			myStartTrigger.Update();
			myValueTrigger.Update();
			myEndTrigger.Update();
			
			if (myClock < 0.0f)
				return;

			myClock -= Time.deltaTime;

			if (myClock <= 0.0f)
			{
				SetValue(1.0f);
				myEndTrigger.Trigger();
				myClock = -1.0f;
			}
			else if (myDuration > 0.001f)
			{
				float v = Mathf.InverseLerp(myDuration, 0.0f, myClock);
				SetValue(Mathf.Clamp01(v));
			}
		}

		private void Trigger()
		{
			if (!myAcceptTrigger.val)
				return;
			
			myClock = myDuration = UnityEngine.Random.Range(myMinTime.val, myMaxTime.val);
			myStartTrigger.Trigger();
			SetValue(0.0f);
		}

		private void SetValue(float v)
		{
			if (myBlendInOut.val)
				v = (v <= 0.5f ? 2.0f*v : 2.0f-2.0f*v);
			if (myBlendSmooth.val)
				v = Mathf.SmoothStep(0.0f, 1.0f, v);

			myValueTrigger.Trigger(v);
		}
    }
}
