﻿using UnityEngine;
using System.Collections.Generic;
using SimpleJSON;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using MeshVR;

namespace MacGruber
{
    public class RandomSoundFromAB : MVRScript
    {
		private JSONStorableUrl myAudioBundle;
		private JSONStorableStringChooser mySoundChooser;
		private JSONStorableBool myAvoidRepetition;
		private JSONStorableBool myAcceptTrigger;
		private JSONStorableStringChooser myModeChooser;
		private JSONStorableString mySoundName;
		private JSONStorableFloat mySoundProbability;

		private string myBundleURL = null;
		private Request myBundleRequest = null;
		private AudioSourceControl myAudioSource;
		private List<Sound> mySounds = new List<Sound>();
		private int myPreviousIndex1 = int.MaxValue;
		private int myPreviousIndex2 = int.MaxValue;
		private bool myNeedUIRefresh = false;
		private List<object> mySoundsUI = new List<object>();
		
		private int myModeIndex = DEFAULT_MODE;		
		private static readonly List<string> MODE_NAMES = new List<string>() { "PlayNext", "PlayNextClearQueue", "QueueClip", "PlayNow", "PlayIfClear", "PlayNowLoop", "PlayNowClearQueue" };
		private const int DEFAULT_MODE = 6;

		private class Sound
		{
			public Sound(string uid)
			{
				this.uid = uid;				
				int start = uid.StartsWith("assets/") ? 7 : 0;
				int end = uid.LastIndexOf('/');
				if (end >= 0)
					name = uid.Substring(start, end-start+1) + "\n" + uid.Substring(end+1);
				else
					name = uid.Substring(start);
			}
			
			public static int SortByNameAscending(Sound a, Sound b)
			{
				return a.uid.CompareTo(b.uid);
			}

			public NamedAudioClip clip;
			public string uid;
			public string name;
		}		

        public override void Init()
		{
			Utils.OnInitUI(CreateUIElement);
						
			string storableName = string.Empty;
			if (containingAtom.type == "Person")
				storableName = "HeadAudioSource";
			else if (containingAtom.type == "AudioSource")
				storableName = "AudioSource";
			else if (containingAtom.type == "AptSpeaker")
				storableName = "AptSpeaker_Import";
			else if (containingAtom.type == "RhythmAudioSource")
				storableName = "RhythmSource";
			myAudioSource = containingAtom.GetStorableByID(storableName) as AudioSourceControl;
			
			if (myAudioSource == null)
			{
				SuperController.LogError("The RandomSoundFromAB plugin needs to be placed on a Person, AudioSource, RhythmAudioSource or AptSpeaker atom.");
				Utils.SetupInfoTextNoScroll(this,
					"<color=#FF6060><size=40><b>RandomSoundFromAB</b></size>\nThis plugin needs to be placed on a Person, AudioSource, RhythmAudioSource or AptSpeaker atom.</color>",
					500.0f, true
				);
				return;
			}

			Utils.SetupInfoTextNoScroll(this,
				"<color=#606060><size=40><b>RandomSoundFromAB</b></size>\nLogicBrick that, after <i>Trigger</i> is called, selects a random <i>Sound</i> from the list to play. Sounds can be added from the selected <i>Unity AssetBundle</i> containing audio files.\n\n" +
				"Use <i>Avoid Repetition</i> take the last one or two choices out of the selection, depending on the number of total available choices.\n\n" +
				"Plugin needs to be placed on a Person, AudioSource, RhythmAudioSource or AptSpeaker atom.</color>",
				500.0f, true
			);
			
			myAudioBundle = Utils.SetupAssetBundleChooser(this, "AssetBundle", "", false, "assetbundle|audiobundle|voicebundle");
			myAudioBundle.setCallbackFunction += (string url) => {
				if (myBundleURL != null)
					AssetLoader.DoneWithAssetBundleFromFile(myBundleURL);
				myBundleURL = url;
				
				//SuperController.LogMessage(myBundleURL);
				if (string.IsNullOrEmpty(myBundleURL))
				{
					OnAssetBundleLoaded(null);
				}
				else
				{
					try {
					Request request = new AssetLoader.AssetBundleFromFileRequest {path = myBundleURL, callback = OnAssetBundleLoaded};
					AssetLoader.QueueLoadAssetBundleFromFile(request);
					}
					catch (System.Exception e)
					{ SuperController.LogError(e.ToString()); }
				}
			};
			
			mySoundChooser = new JSONStorableStringChooser("Sound", new List<string>(), "", "");
			UIDynamicPopup popup = CreateFilterablePopup(mySoundChooser, false);
			popup.popupPanelHeight = 500;
			popup.labelWidth = -8;
			
			RegisterStringChooser(mySoundChooser);
			mySoundChooser.isStorable = mySoundChooser.isRestorable = false;
			
			Utils.SetupTwinButton(this, "Play Sound", UIPlaySound, "Stop Sound", UIStopSound, false);
			Utils.SetupTwinButton(this, "Add Sound", UIAddSound, "Add All", UIAddAllSound, false);
			myAvoidRepetition = Utils.SetupToggle(this, "Avoid Repetition", true, false);
			myAcceptTrigger = Utils.SetupToggle(this, "AcceptTrigger", true, false);
			myModeChooser = Utils.SetupStringChooser(this, "Mode", MODE_NAMES, DEFAULT_MODE, false);
			myModeChooser.setCallbackFunction += (string v) => {
				myModeIndex = MODE_NAMES.FindIndex((string entry) => { return entry == v; });
				if (myModeIndex < 0)
				{
					myModeIndex = DEFAULT_MODE;
					myModeChooser.valNoCallback = MODE_NAMES[myModeIndex];
				}
			};

			Utils.SetupButton(this, "Test Trigger", Trigger, false);
			Utils.SetupAction(this, "Trigger", Trigger);

			Utils.SetupSpacer(this, 15, false);			
			
			UIUpdate();
		}
		
		private void OnAssetBundleLoaded(Request aRequest)
		{
			try {
				myBundleRequest = aRequest;
				List<string> names = new List<string>();
				List<string> uids = new List<string>();			
				if (myBundleRequest != null)
				{
					string[] assets = myBundleRequest.assetBundle?.GetAllAssetNames();					
					for (int i=0; i<assets.Length; ++i)
					{
						string uid = assets[i];				
						if (uid.EndsWith(".wav") || uid.EndsWith(".ogg") || uid.EndsWith(".mp3") || uid.EndsWith(".aif"))
							uids.Add(uid);
					}					
					uids.Sort();
					for (int i=0; i<uids.Count; ++i)
					{
						string uid = uids[i];
						int start = uid.StartsWith("assets/") ? 7 : 0;
						int end = uid.LastIndexOf('/');
						if (end >= 0)
							names.Add(uid.Substring(start, end-start+1) + "\n" + uid.Substring(end+1));
						else
							names.Add(uid.Substring(start));
					}
				}
				mySoundChooser.choices = uids;
				mySoundChooser.displayChoices = names;
				
				if (uids.Count == 0)
					mySoundChooser.val = "";
				else if (!uids.Contains(mySoundChooser.val))
					mySoundChooser.val = uids[0];
				
				for (int i=0; i<mySounds.Count; ++i)
				{
					Sound sound = mySounds[i];
					sound.clip = null;
					if (uids.Contains(sound.uid))
					{
						AssetBundleAudioClip clip = new AssetBundleAudioClip(myBundleRequest, "", sound.uid);
						if (clip.sourceClip != null)
							sound.clip = clip;
					}
				}
				mySounds.RemoveAll(s => s.clip == null);
				
				UIUpdate();
			}
			catch (System.Exception e)
			{ SuperController.LogError(e.ToString()); }
		}

		private void UIUpdate()
		{
			Utils.RemoveUIElements(this, mySoundsUI);
			
			if (mySounds.Count > 0)
			{
				UIDynamicTextInfo header = Utils.SetupInfoOneLine(this, "<size=30><b>Sounds</b></size>", false);
				mySoundsUI.Add(header);
				for (int i=0; i<mySounds.Count; ++i)
				{
					Sound sound = mySounds[i];
					UIDynamicLabelXButton label = Utils.SetupLabelXButton(this, sound.name, () => UIRemoveSound(sound), false);
					mySoundsUI.Add(label);
				}
			}
		}
		
		private NamedAudioClip UIGetClip(string uid)
		{
			if (string.IsNullOrEmpty(uid))
				return null;
			AssetBundleAudioClip clip = new AssetBundleAudioClip(myBundleRequest, "", uid);
			if (clip.sourceClip == null)
				return null;
			return clip;
		}
		
		private void UIPlaySound()
		{
			string uid = mySoundChooser.val;
			NamedAudioClip clip = UIGetClip(uid);
			if (clip != null && myAudioSource != null)
				myAudioSource.PlayNowClearQueue(clip);
		}
		
		private void UIStopSound()
		{
			if (myAudioSource != null)
				myAudioSource.StopAndClearQueue();
		}

		private void UIAddSound()
		{
			string uid = mySoundChooser.val;
			NamedAudioClip clip = UIGetClip(uid);
			if (clip != null)
			{
				Sound sound = new Sound(uid);
				sound.clip = clip;
				mySounds.Add(sound);
				mySounds.Sort(Sound.SortByNameAscending);
				UIUpdate();
			}
		}
		
		private void UIAddAllSound()
		{
			mySounds.Clear();
			
			List<string> uids = mySoundChooser.choices;
			for (int i=0; i<uids.Count; ++i)
			{
				NamedAudioClip clip = UIGetClip(uids[i]);
				if (clip == null)
					continue;
					
				Sound sound = new Sound(uids[i]);
				sound.clip = clip;
				mySounds.Add(sound);
			}
			
			mySounds.Sort(Sound.SortByNameAscending);
			UIUpdate();			
		}

		private void UIRemoveSound(Sound sound)
		{
			mySounds.Remove(sound);
			UIUpdate();
		}
		
		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if (includePhysical || forceStore)
			{
				needsStore = true;

				JSONArray slist = new JSONArray();
				for (int i=0; i<mySounds.Count; ++i)
					slist.Add("", mySounds[i].uid);
				jc["Sounds"] = slist;
			}

			return jc;
		}

		public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
		{
			base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
			if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
			{
				mySounds.Clear();
 				JSONArray slist = jc["Sounds"].AsArray;
				for (int i=0; i<slist.Count; ++i)
				{
					string uid = slist[i].Value;
					mySounds.Add(new Sound(uid));
				}
				mySounds.Sort(Sound.SortByNameAscending);
				myNeedUIRefresh = true;
			}
		}

		private void Update()
		{
			if (myBundleURL == null)
				myAudioBundle.setCallbackFunction(myAudioBundle.val);
			
			if (myNeedUIRefresh)
			{
				myNeedUIRefresh = false;
				UIUpdate();
			}
		}

		private void OnDestroy()
		{
			mySounds.Clear();
			Utils.OnDestroyUI();
			
			if (myBundleURL != null)
                AssetLoader.DoneWithAssetBundleFromFile(myBundleURL);
			myBundleURL = null;
		}

		private void Trigger()
		{
			int count = mySounds.Count;
			if (count == 0 || !myAcceptTrigger.val)
				return;

			if (count < 3 || myPreviousIndex1 >= count)
				myPreviousIndex1 = int.MaxValue;
			if (count < 4 || myPreviousIndex2 >= count)
				myPreviousIndex2 = int.MaxValue;
			if (myPreviousIndex1 == myPreviousIndex2)
				myPreviousIndex2 = int.MaxValue;
			if (myAvoidRepetition.val && myPreviousIndex1 < int.MaxValue)
				--count;
			if (myAvoidRepetition.val && myPreviousIndex2 < int.MaxValue)
				--count;
						
			int idx = UnityEngine.Random.Range(0, count);			
			if (idx >= Mathf.Min(myPreviousIndex1, myPreviousIndex2))
				++idx;
			if (idx >= Mathf.Max(myPreviousIndex1, myPreviousIndex2))
				++idx;
					
			myPreviousIndex2 = myPreviousIndex1;
			myPreviousIndex1 = idx;
			
			NamedAudioClip clip = mySounds[idx].clip;
			if (myAudioSource == null || clip == null)
				return;
			
			switch (myModeIndex)
			{
				case 0: myAudioSource.PlayNext(clip);           break;
				case 1: myAudioSource.PlayNextClearQueue(clip); break;
				case 2: myAudioSource.QueueClip(clip);          break;
				case 3: myAudioSource.PlayNow(clip);            break;
				case 4: myAudioSource.PlayIfClear(clip);        break;
				case 5: myAudioSource.PlayNowLoop(clip);        break;
				case 6: myAudioSource.PlayNowClearQueue(clip);  break;
				default: break;
			}
		}
    }
}
