﻿using UnityEngine;
using System.Collections.Generic;
using SimpleJSON;

namespace MacGruber
{
    public class SelectChoice : MVRScript
    {
		private JSONStorableString myCurrentChoiceInfo;
		private JSONStorableStringChooser myChoiceChooser;
		private JSONStorableBool myAcceptTrigger;
		private JSONStorableString myChoiceName;
		private JSONStorableString mySelectChoice;

		private List<object> myChoiceUI = new List<object>();

		private class Choice
		{
			public Choice(MVRScript script, string name)
			{
				myName = name;
				Trigger = new EventTrigger(script, "On" + name);
			}
			
			public Choice(MVRScript script, string name, Choice source)
			{
				Trigger = new EventTrigger(source.Trigger);
				Name = name;
			}

			public static int SortByNameAscending(Choice a, Choice b)
			{
				return a.Name.CompareTo(b.Name);
			}

			public string Name {
				get { return myName; }
				set {
					myName = value;
					Trigger.Name = "On" + value;
				}
			}

			public EventTrigger Trigger;
			private string myName;
		}

		private List<Choice> myChoices = new List<Choice>();
		private int mySelectedIndex = -1;
		private bool myNeedUIRefresh = false;

        public override void Init()
		{
			Utils.OnInitUI(CreateUIElement);

			Utils.SetupInfoText(this,
				"<color=#606060><size=40><b>SelectChoice</b></size>\nLogicBrick that, when <i>Trigger</i> is called, executes a selected <i>Choice</i>. Use the " +
				"<i>SelectChoice</i> trigger with the name of your <i>Choice</i>, or when debugging the button on the left.\n\nFor coders: This brick is the equivalent of a <i>switch</i> statement.</color>",
				320.0f, true
			);
			
			myCurrentChoiceInfo = new JSONStorableString("CurrentChoice", "Current Choice:");	
			Utils.SetupInfoOneLine(this, myCurrentChoiceInfo, true);
			
			myChoiceChooser = new JSONStorableStringChooser("Choice", new List<string>(), "", "Choice");		
			myChoiceChooser.isStorable = myChoiceChooser.isRestorable = false;
			myChoiceChooser.setCallbackFunction += UISetChoice;
			CreateScrollablePopup(myChoiceChooser, false);

			Utils.SetupTwinButton(this, "Add Choice", UIAddChoice, "Duplicate Choice", UIDuplicateChoice, false);
			Utils.SetupButton(this, "Remove Choice", UIRemoveChoice, false);
			
			myAcceptTrigger = Utils.SetupToggle(this, "AcceptTrigger", true, false);

			Utils.SetupButton(this, "Select Choice", SelectChoiceButton, false);
			Utils.SetupButton(this, "Test Trigger", TriggerAction, false);
			
			mySelectChoice = new JSONStorableString("SelectChoice", "", SelectChoiceAction);
			mySelectChoice.isStorable = mySelectChoice.isRestorable = false;
			RegisterString(mySelectChoice);
			Utils.SetupAction(this, "Trigger", TriggerAction);			

			Utils.SetupSpacer(this, 15, false);

			myChoiceName = new JSONStorableString("Name", "", UIRenameChoice);

			SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
			SimpleTriggerHandler.LoadAssets();
		}
		
		private void SelectChoiceButton()
		{			
			mySelectedIndex = myChoiceChooser.choices.IndexOf(myChoiceChooser.val);			
			if (mySelectedIndex >= 0)
				myCurrentChoiceInfo.val = "Current Choice: " + myChoiceChooser.val;
			else
				myCurrentChoiceInfo.val = "Current Choice:";
		}
		
		private void SelectChoiceAction(string newChoice)
		{
			mySelectChoice.valNoCallback = string.Empty;
			int selectedIndex = myChoiceChooser.choices.IndexOf(newChoice);
			if (selectedIndex >= 0)
			{
				mySelectedIndex = selectedIndex;
				myCurrentChoiceInfo.val = "Current Choice: " + newChoice;
			}
			else
			{
				myCurrentChoiceInfo.val = "Current Choice:";
				SuperController.LogError($"LogicBricks.SelectChoice: Unknown choice '{newChoice}'.");
			}
		}
		
		private void UISetChoice(string v)
		{
			Utils.RemoveUIElements(this, myChoiceUI);

			int idx = myChoiceChooser.choices.IndexOf(v);
			if (idx < 0)
				return;

			Choice choice = myChoices[idx];

			UIDynamicTextInfo optionHeader = Utils.SetupInfoOneLine(this, "<size=30><b>Choice</b></size>", false);
			myChoiceUI.Add(optionHeader);

			myChoiceName.valNoCallback = choice.Name;
			UIDynamicLabelInput nameInput = Utils.SetupTextInput(this, "Name", myChoiceName, false);
			myChoiceUI.Add(nameInput);			

			UIDynamicButton actionsButton = Utils.SetupButton(this, "Actions " + choice.Trigger.Name, choice.Trigger.OpenPanel, false);
			myChoiceUI.Add(actionsButton);
		}
		
		private void UIAddChoice()
		{
			UIAddChoice(null);
		}

		private void UIAddChoice(Choice source)
		{
			for (int i=1; i<1000; ++i)
			{
				string name = "Choice" + i;
				if (!myChoiceChooser.choices.Contains(name))
				{
					Choice choice;
					if (source == null)
						choice = new Choice(this, name);
					else
						choice = new Choice(this, name, source);
					myChoices.Add(choice);
					myChoices.Sort(Choice.SortByNameAscending);
					List<string> options = myChoiceChooser.choices;
					options.Clear();
					for (int e=0; e<myChoices.Count; ++e)
						options.Add(myChoices[e].Name);
					myChoiceChooser.choices = null; // force UI sync
					myChoiceChooser.choices = options;
					myChoiceChooser.val = name;
					return;
				}
			}
			SuperController.LogError("RandomChoice: Too many options!");
		}
		
		private void UIDuplicateChoice()
		{
			int idx = myChoiceChooser.choices.IndexOf(myChoiceChooser.val);
			Choice choice = idx >= 0 ? myChoices[idx] : null;
			UIAddChoice(choice);
		}

		private void UIRemoveChoice()
		{
			int idx = myChoiceChooser.choices.IndexOf(myChoiceChooser.val);
			if (idx < 0)
				return;

			List<string> options = myChoiceChooser.choices;
			options.RemoveAt(idx);
			myChoices[idx].Trigger.Remove();
			myChoices.RemoveAt(idx);
			myChoiceChooser.choices = null; // force UI sync
			myChoiceChooser.choices = options;

			if (idx < options.Count)
				myChoiceChooser.val = options[idx];
			else if (options.Count > 0)
				myChoiceChooser.val = options[options.Count-1];
			else
				myChoiceChooser.val = "";
		}

		private void UIRenameChoice(string name)
		{
			int idx = myChoiceChooser.choices.IndexOf(myChoiceChooser.val);
			if (idx < 0)
				return;

			myChoices[idx].Name = name;

			myChoices.Sort(Choice.SortByNameAscending);
			List<string> options = myChoiceChooser.choices;
			options.Clear();
			for (int e=0; e<myChoices.Count; ++e)
				options.Add(myChoices[e].Name);
			myChoiceChooser.choices = null; // force UI sync
			myChoiceChooser.choices = options;
			myChoiceChooser.val = name;
		}

		private void OnAtomRename(string oldid, string newid)
		{
			for (int i=0; i<myChoices.Count; ++i)
				myChoices[i].Trigger.SyncAtomNames();
		}

		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if (includePhysical || forceStore)
			{
				needsStore = true;

				JSONArray clist = new JSONArray();
				for (int i=0; i<myChoices.Count; ++i)
				{
					Choice choice = myChoices[i];
					JSONClass cc = new JSONClass();
					cc["Name"] = choice.Name;
					cc[choice.Trigger.Name] = choice.Trigger.GetJSON(base.subScenePrefix);
					clist.Add("", cc);
				}
				jc["Choices"] = clist;
				
				if (mySelectedIndex >= 0 && mySelectedIndex < myChoices.Count)
					jc["Selected"] = myChoices[mySelectedIndex].Name;
			}

			return jc;
		}

		public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
		{
			base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
			if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
			{
				for (int i=0; i<myChoices.Count; ++i)
					myChoices[i].Trigger.Remove();
				myChoices.Clear();

				JSONArray clist = jc["Choices"].AsArray;
				for (int i=0; i<clist.Count; ++i)
				{
					JSONClass cc = clist[i].AsObject;
					Choice choice = new Choice(this, cc["Name"]);
					choice.Trigger.RestoreFromJSON(cc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
					myChoices.Add(choice);
				}

				myChoices.Sort(Choice.SortByNameAscending);
				List<string> options = myChoiceChooser.choices;
				options.Clear();
				for (int e=0; e<myChoices.Count; ++e)
					options.Add(myChoices[e].Name);
				myChoiceChooser.choices = null; // force UI sync
				myChoiceChooser.choices = options;

				if (options.Count == 0)
					myChoiceChooser.valNoCallback = "";
				else if (!options.Contains(myChoiceChooser.val))
					myChoiceChooser.valNoCallback = options[0];
				
				mySelectedIndex = -1;
				if (jc.HasKey("Selected"))
					SelectChoiceAction(jc["Selected"]);
				
				myNeedUIRefresh = true;
			}
		}

		private void Update()
		{
			for (int i=0; i<myChoices.Count; ++i)
				myChoices[i].Trigger.Update();
			
			if (myNeedUIRefresh)
			{
				myNeedUIRefresh = false;
				myChoiceChooser.setCallbackFunction(myChoiceChooser.val);
			}
		}

		private void OnDestroy()
		{
			SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;

			for (int i=0; i<myChoices.Count; ++i)
				myChoices[i].Trigger.Remove();
			myChoices.Clear();

			Utils.OnDestroyUI();
		}

		private void TriggerAction()
		{
			if (myAcceptTrigger.val && mySelectedIndex >= 0 && mySelectedIndex < myChoices.Count)
				myChoices[mySelectedIndex].Trigger.Trigger();
		}
    }
}
