﻿using UnityEngine;
using UnityEngine.UI;
using System.Collections.Generic;
using SimpleJSON;

namespace MacGruber
{
    public class StateMachine : MVRScript
    {
		private JSONStorableStringChooser myStateChooser;
		private JSONStorableString myCurrentStateInfo;
		private JSONStorableString myCurrentClockInfo;
		private JSONStorableString mySwitchState;
		private JSONStorableString mySetOnCooldown;
		private JSONStorableString mySetEnabled;
		private JSONStorableString mySetDisabled;
		private JSONStorableBool myPaused;
		private JSONStorableBool myShowAdvancedOptions;
		private Text myShowAdvancedOptionsLabel;
		private JSONStorableString myStateName;
		private JSONStorableBool myStateIsEnabled;
		private JSONStorableBool myStateWaitForSync;
		private JSONStorableBool myStateInfiniteDuration;
		private JSONStorableFloat myStateMinDuration;
		private JSONStorableFloat myStateMaxDuration;
		private JSONStorableFloat myStateCooldown;
		private JSONStorableString myStateTransitions;
		private JSONStorableStringChooser myChoiceChooser;
		private JSONStorableStringChooser myTransitionChooser;		

		private List<object> myStateUI = new List<object>();

		private State myCurrentState = null;
		private int myCurrentChoice = -1;
		private float myClock = 0.0f;
		private float myDuration = 1.0f;
		private float myValue = -1.0f;
		private int myClockInt = -1;
		private bool myNeedUIRefresh = false;
		private bool myNeedUIRebuild = false;
		
		private static readonly List<string> CHOICE_NAMES = new List<string>() { "Default Timeout", "Choice A", "Choice B", "Choice C", "Choice D", "Choice E", "Choice F" };

		private class State
		{
			public State(MVRScript script, string name)
			{
				myName = name;
				EnterTrigger = new EventTrigger(script, "OnEnter" + name);
				ValueTrigger = new FloatTrigger(script, "OnValue" + name);
				ExitTrigger = new EventTrigger(script, "OnExit" + name);
				
				ChoiceTransitions = new List<List<State>>(CHOICE_NAMES.Count);
				for (int i=0; i<CHOICE_NAMES.Count; ++i)
					ChoiceTransitions.Add(new List<State>(0));
			}

			public State(MVRScript script, string name, State source)
			{
				EnterTrigger = new EventTrigger(source.EnterTrigger);
				ValueTrigger = new FloatTrigger(source.ValueTrigger);
				ExitTrigger = new EventTrigger(source.ExitTrigger);
				Name = name;
				Enabled = source.Enabled;
				InfiniteDuration = source.InfiniteDuration;
				WaitForSync = source.WaitForSync;
				MinDuration = source.MinDuration;
				MaxDuration = source.MaxDuration;
				Cooldown = source.Cooldown;
				Timestamp = source.Timestamp;
				
				ChoiceTransitions = new List<List<State>>(CHOICE_NAMES.Count);
				for (int i=0; i<CHOICE_NAMES.Count; ++i)
					ChoiceTransitions.Add(new List<State>(source.ChoiceTransitions[i]));
			}

			public static int SortByNameAscending(State a, State b)
			{
				return a.Name.CompareTo(b.Name);
			}

			public string Name {
				get { return myName; }
				set {
					myName = value;
					EnterTrigger.Name = "OnEnter" + value;
					ValueTrigger.Name = "OnValue" + value;
					ExitTrigger.Name = "OnExit" + value;
				}
			}

			public float RemainingCooldown {
				get { return Cooldown - Time.time + Timestamp; }
			}

			public EventTrigger EnterTrigger;
			public FloatTrigger ValueTrigger;
			public EventTrigger ExitTrigger;
			public bool Enabled = true;
			public bool InfiniteDuration = false;
			public bool WaitForSync = false;
			public float MinDuration = 3.0f;
			public float MaxDuration = 5.0f;
			public float Cooldown = 0.0f;
			public float Timestamp = -10000.0f;
			public int UISelectedChoice = 0; // not saved		
			public List<List<State>> ChoiceTransitions;

			private string myName;			
		}

		private List<State> myStates = new List<State>();

        public override void Init()
		{
			Utils.OnInitUI(CreateUIElement);

			Utils.SetupInfoText(this,
				"<color=#606060><size=40><b>StateMachine</b></size>\nPowerful LogicBrick that represents the concept of a StateMachine.</color>",
				120.0f, true
			);

			myCurrentStateInfo = new JSONStorableString("CurrentState", "Current State:");
			myCurrentClockInfo = new JSONStorableString("CurrentClock", "Current Clock:");
			Utils.SetupInfoOneLine(this, myCurrentStateInfo, true);
			Utils.SetupInfoOneLine(this, myCurrentClockInfo, true);

			myStateChooser = new JSONStorableStringChooser("Selected State", new List<string>(), string.Empty, "Selected State");
			myStateChooser.displayChoices = new List<string>();
			myStateChooser.setCallbackFunction += UISetState;
			CreateFilterablePopup(myStateChooser, false);

			Utils.SetupTwinButton(this, "Add State", UIAddState, "Duplicate State", UIDuplicateState, false);
			Utils.SetupButton(this, "Remove State", UIRemoveState, false);
			
			Utils.SetupSpacer(this, 5, false);

			
			Utils.SetupTwinButton(this, "Switch State", UISwitchState, "Skip State", UISkipState, false);

			Utils.SetupAction(this, "TriggerSync", TriggerSyncAction);

			mySwitchState = new JSONStorableString("SwitchState", "", SwitchStateAction);
			mySwitchState.isStorable = mySwitchState.isRestorable = false;
			RegisterString(mySwitchState);

			mySetOnCooldown = new JSONStorableString("SetOnCooldown", "", SetOnCooldownAction);
			mySetOnCooldown.isStorable = mySetOnCooldown.isRestorable = false;
			RegisterString(mySetOnCooldown);

			mySetEnabled = new JSONStorableString("SetStateEnabled", "", SetEnabledAction);
			mySetEnabled.isStorable = mySetEnabled.isRestorable = false;
			RegisterString(mySetEnabled);			
			Utils.SetupAction(this, "SetStateEnabledAll", () => SetEnabledAllAction(true));

			mySetDisabled = new JSONStorableString("SetStateDisabled", "", SetDisabledAction);
			mySetDisabled.isStorable = mySetDisabled.isRestorable = false;
			RegisterString(mySetDisabled);			
			Utils.SetupAction(this, "SetStateDisabledAll", () => SetEnabledAllAction(false));
			
			Utils.SetupAction(this, "SetChoiceDefault", () => SetChoiceAction(0));
			Utils.SetupAction(this, "SetChoiceA", () => SetChoiceAction(1));
			Utils.SetupAction(this, "SetChoiceB", () => SetChoiceAction(2));
			Utils.SetupAction(this, "SetChoiceC", () => SetChoiceAction(3));
			Utils.SetupAction(this, "SetChoiceD", () => SetChoiceAction(4));
			Utils.SetupAction(this, "SetChoiceE", () => SetChoiceAction(5));
			Utils.SetupAction(this, "SetChoiceF", () => SetChoiceAction(6));
			Utils.SetupAction(this, "SetChoiceA and Wait", () => SetChoiceAction(-2));
			Utils.SetupAction(this, "SetChoiceB and Wait", () => SetChoiceAction(-3));
			Utils.SetupAction(this, "SetChoiceC and Wait", () => SetChoiceAction(-4));
			Utils.SetupAction(this, "SetChoiceD and Wait", () => SetChoiceAction(-5));
			Utils.SetupAction(this, "SetChoiceE and Wait", () => SetChoiceAction(-6));
			Utils.SetupAction(this, "SetChoiceF and Wait", () => SetChoiceAction(-7));

			myPaused = new JSONStorableBool("Time Paused (not saved)", false);
			CreateToggle(myPaused, false);
			
			myShowAdvancedOptions = new JSONStorableBool("Show Advanced Options", false);
			myShowAdvancedOptions.setCallbackFunction += UISetAdvancedOptions;
			UIDynamicToggle advancedToggle = CreateToggle(myShowAdvancedOptions, false);
			myShowAdvancedOptionsLabel = advancedToggle.labelText;
			
			myStateName = new JSONStorableString("Name", "", UIRenameState);
			myStateIsEnabled = new JSONStorableBool("Enabled for Transitions", false);
			myStateIsEnabled.setCallbackFunction += UISetStateEnabled;
			myStateInfiniteDuration = new JSONStorableBool("Infinite Duration", false);
			myStateInfiniteDuration.setCallbackFunction += UISetInfiniteDuration;
			myStateWaitForSync = new JSONStorableBool("Wait for TriggerSync", false);
			myStateWaitForSync.setCallbackFunction += UISetWaitForSync;
			myStateMinDuration = new JSONStorableFloat("Min Duration (sec)", 3, 0, 10, true, true);
			myStateMinDuration.setCallbackFunction += UISetMinDuration;
			myStateMinDuration.constrained = false;
			myStateMaxDuration = new JSONStorableFloat("Max Duration (sec)", 5, 0, 10, true, true);
			myStateMaxDuration.setCallbackFunction += UISetMaxDuration;
			myStateMaxDuration.constrained = false;
			myStateCooldown = new JSONStorableFloat("Cooldown (sec)", 5, 0, 10, true, true);
			myStateCooldown.setCallbackFunction += UISetCooldown;
			myStateCooldown.constrained = false;
			
			myChoiceChooser = new JSONStorableStringChooser("Selected Choice", CHOICE_NAMES, CHOICE_NAMES[0], "Selected Choice");
			myChoiceChooser.setCallbackFunction += UISetChoice;

			myTransitionChooser = new JSONStorableStringChooser("Transition", new List<string>(), "", "Transition to Add");
			myTransitionChooser.displayChoices = new List<string>();

			SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
			SimpleTriggerHandler.LoadAssets();
		}

		private void UIRefresh()
		{
			myNeedUIRefresh = false;
			UISetState(myStateChooser.val);
		}

		private void UIRebuild()
		{
			myNeedUIRebuild = false;
			List<string> options = myStateChooser.choices;
			List<string> displays = myStateChooser.displayChoices;
			options.Clear();
			displays.Clear();
			for (int e=0; e<myStates.Count; ++e)
			{
				State state = myStates[e];
				options.Add(state.Name);
				displays.Add(state.Enabled ? state.Name : $"<color=#CC0000>{state.Name}</color>");
			}
			myStateChooser.choices = null; // force UI sync
			myStateChooser.choices = options;

			string current = myStateChooser.val;
			myStateChooser.valNoCallback = "";
			myStateChooser.valNoCallback = current;
		}

		private void UISetState(string v)
		{
			Utils.RemoveUIElements(this, myStateUI);

			State state = GetSelectedState();
			if (state == null)
				return;		
		
			// This state has advanced options set, but we are not showing them?
			bool hasAdvancedOptions = !state.Enabled || state.WaitForSync || state.Cooldown > 0;
			for (int t=1; t<state.ChoiceTransitions.Count; ++t)
				hasAdvancedOptions |= state.ChoiceTransitions[t].Count > 0;			
			myShowAdvancedOptionsLabel.text = !myShowAdvancedOptions.val && hasAdvancedOptions ? "<color=#CC0000><b>Show Advanced Options</b></color>" : "Show Advanced Options";

			UIDynamicTextInfo stateHeader = Utils.SetupInfoOneLine(this, "<size=30><b>State</b></size>", false);
			myStateUI.Add(stateHeader);

			myStateName.valNoCallback = state.Name;
			UIDynamicLabelInput nameInput = Utils.SetupTextInput(this, "Name", myStateName, false);
			myStateUI.Add(nameInput);

			if (myShowAdvancedOptions.val)
			{
				myStateIsEnabled.valNoCallback = state.Enabled;
				CreateToggle(myStateIsEnabled, false);
				myStateUI.Add(myStateIsEnabled);
			}

			myStateInfiniteDuration.valNoCallback = state.InfiniteDuration;
			CreateToggle(myStateInfiniteDuration, false);
			myStateUI.Add(myStateInfiniteDuration);

			if (!state.InfiniteDuration)
			{
				if (myShowAdvancedOptions.val)
				{
					myStateWaitForSync.valNoCallback = state.WaitForSync;
					CreateToggle(myStateWaitForSync, false);
					myStateUI.Add(myStateWaitForSync);
				}

				myStateMinDuration.valNoCallback = state.MinDuration;
				myStateMaxDuration.valNoCallback = state.MaxDuration;
				Utils.AdjustSliderRange(myStateMinDuration);
				Utils.AdjustSliderRange(myStateMaxDuration);
				UIDynamicSlider sliderMin = CreateSlider(myStateMinDuration, false);
				UIDynamicSlider sliderMax = CreateSlider(myStateMaxDuration, false);
				sliderMin.rangeAdjustEnabled = true;
				sliderMax.rangeAdjustEnabled = true;
				myStateUI.Add(myStateMinDuration);
				myStateUI.Add(myStateMaxDuration);
			}

			if (myShowAdvancedOptions.val)
			{
				myStateCooldown.valNoCallback = state.Cooldown;
				Utils.AdjustSliderRange(myStateCooldown);
				UIDynamicSlider sliderCooldown = CreateSlider(myStateCooldown, false);
				sliderCooldown.rangeAdjustEnabled = true;
				myStateUI.Add(myStateCooldown);
			}

			UIDynamicTextInfo triggersHeader = Utils.SetupInfoOneLine(this, "<size=30><b>Triggers</b></size>", true);
			myStateUI.Add(triggersHeader);
			UIDynamicButton enterButton = Utils.SetupButton(this, "Actions " + state.EnterTrigger.Name, state.EnterTrigger.OpenPanel, true);
			UIDynamicButton valueButton = Utils.SetupButton(this, "Actions " + state.ValueTrigger.Name, state.ValueTrigger.OpenPanel, true);
			UIDynamicButton exitButton = Utils.SetupButton(this, "Actions " + state.ExitTrigger.Name, state.ExitTrigger.OpenPanel, true);
			myStateUI.Add(enterButton);
			myStateUI.Add(valueButton);
			myStateUI.Add(exitButton);

			if (myStates.Count > 0)
			{
				UIDynamicTextInfo transitionsHeader = Utils.SetupInfoOneLine(this, "<size=30><b>Transitions</b></size>", true);
				myStateUI.Add(transitionsHeader);
				
				if (myShowAdvancedOptions.val)
				{
					myChoiceChooser.valNoCallback = CHOICE_NAMES[state.UISelectedChoice];
					UIDynamicPopup choicePopup = CreateScrollablePopup(myChoiceChooser, true);
					myStateUI.Add(choicePopup);
				}
				else
				{
					state.UISelectedChoice = 0;
				}

				List<string> transitionNames = myTransitionChooser.choices;
				List<string> displayNames = myTransitionChooser.displayChoices;
				transitionNames.Clear();
				displayNames.Clear();
				for (int s=0; s<myStates.Count; ++s)
				{
					State transition = myStates[s];
					transitionNames.Add(transition.Name);
					displayNames.Add(transition.Enabled ? transition.Name : $"<color=#CC0000>{transition.Name}</color>");
				}
				myTransitionChooser.choices = null;
				myTransitionChooser.choices = transitionNames;
				string currentTransition = myTransitionChooser.val;
				if (transitionNames.Count == 0)
				{
					myTransitionChooser.val = "";
				}
				else if (!transitionNames.Contains(currentTransition))
				{
					myTransitionChooser.val = transitionNames[0];
				}
				else
				{
					myTransitionChooser.valNoCallback = "";
					myTransitionChooser.valNoCallback = currentTransition;
				}
				UIDynamicPopup transitionPopup = CreateFilterablePopup(myTransitionChooser, true);
				myStateUI.Add(transitionPopup);

				UIDynamicButton transitionButton = Utils.SetupButton(this, "Add Transition", () => UIAddTransition(state), true);
				myStateUI.Add(transitionButton);
				
				List<State> transitions = state.ChoiceTransitions[state.UISelectedChoice];
				for (int t=0; t<transitions.Count; ++t)
				{
					State transition = transitions[t];
					string label = transition.Enabled ? transition.Name : $"<color=#CC0000>{transition.Name}</color>";
					UIDynamicLabelXButton button = Utils.SetupLabelXButton(this, label, () => UIRemoveTransition(state, transition), true);
					myStateUI.Add(button);
				}
			}
		}
		
		private void UISetAdvancedOptions(bool v)
		{			
			UIRefresh();
		}

		private void UISetStateEnabled(bool v)
		{
			State state = GetSelectedState();
			if (state != null && state.Enabled != v)
			{
				state.Enabled = v;
				UIRebuild();
				UIRefresh();
			}
		}

		private void UISetInfiniteDuration(bool v)
		{
			State state = GetSelectedState();
			if (state == null)
				return;

			state.InfiniteDuration = v;
			state.WaitForSync &= !v;
			UIRefresh();

			myClockInt = -1;
			myValue = -1.0f;
			UpdateCurrentStateInfo();
			UpdateCurrentStateValue();
		}

		private void UISetWaitForSync(bool v)
		{
			State state = GetSelectedState();
			if (state == null)
				return;

			state.WaitForSync = v;
		}

		private void UISetMinDuration(float v)
		{
			State state = GetSelectedState();
			if (state == null)
				return;

			state.MinDuration = v;
			Utils.AdjustMaxSliderFromMin(v, myStateMaxDuration);
			state.MaxDuration = myStateMaxDuration.val;
		}

		private void UISetMaxDuration(float v)
		{
			State state = GetSelectedState();
			if (state == null)
				return;

			state.MaxDuration = v;
			state.MinDuration = Mathf.Min(v, state.MinDuration);
			myStateMinDuration.valNoCallback = state.MinDuration;
		}

		private void UISetCooldown(float v)
		{
			State state = GetSelectedState();
			if (state == null)
				return;

			state.Cooldown = v;
		}
		
		private void UISetChoice(string v)
		{
			State state = GetSelectedState();
			if (state == null)
				return;
			
			int idx = CHOICE_NAMES.FindIndex((string entry) => { return entry == v; });
			state.UISelectedChoice = Mathf.Max(idx, 0);
			
			UIRefresh();
		}

		private void UIAddState()
		{
			UIAddState(null);
		}

		private void UIAddState(State source)
		{
			for (int i=1; i<1000; ++i)
			{
				string name = "State" + i;
				if (!myStateChooser.choices.Contains(name))
				{
					State state;
					if (source == null)
						state = new State(this, name);
					else
						state = new State(this, name, source);
					myStates.Add(state);
					myStates.Sort(State.SortByNameAscending);

					UIRebuild();
					myStateChooser.val = name;
					return;
				}
			}
			SuperController.LogError($"LogicBricks.StateMachine on atom '{containingAtom.uid}': Too many states!");
		}

		private void UIDuplicateState()
		{
			int idx = myStateChooser.choices.IndexOf(myStateChooser.val);
			State state = idx >= 0 ? myStates[idx] : null;
			UIAddState(state);
		}

		private void UIRemoveState()
		{
			int idx = myStateChooser.choices.IndexOf(myStateChooser.val);
			if (idx < 0)
				return;

			State state = myStates[idx];
			if (myCurrentState == state)
				SwitchState(null);
			for (int s=0; s<myStates.Count; ++s)
			{
				List<List<State>> choiceTransitions = myStates[s].ChoiceTransitions;
				for (int c=0; c<choiceTransitions.Count; ++c)
					choiceTransitions[c].Remove(state);
			}

			state.EnterTrigger.Remove();
			state.ValueTrigger.Remove();
			state.ExitTrigger.Remove();
			myStates.RemoveAt(idx);

			List<string> options = myStateChooser.choices;
			List<string> displays = myStateChooser.displayChoices;
			options.RemoveAt(idx);
			displays.RemoveAt(idx);
			myStateChooser.choices = null; // force UI sync
			myStateChooser.choices = options;

			if (idx < options.Count)
				myStateChooser.val = options[idx];
			else if (options.Count > 0)
				myStateChooser.val = options[options.Count-1];
			else
				myStateChooser.val = "";
		}

		private void UISwitchState()
		{
			SwitchState(GetSelectedState());
		}
		
		private void UISkipState()
		{
			myClock = 0.0f;
		}

		private void UIRenameState(string name)
		{
			State state = GetSelectedState();
			if (state == null)
				return;

			state.Name = name;
			myStates.Sort(State.SortByNameAscending);

			UIRebuild();
			myStateChooser.val = name;
		}

		private void UIAddTransition(State state)
		{
			State transition = GetState(myTransitionChooser.val);
			if (transition == null)
				return;

			List<State> transitions = state.ChoiceTransitions[state.UISelectedChoice];
			transitions.Add(transition);
			transitions.Sort(State.SortByNameAscending);
			UIRefresh();
		}

		private void UIRemoveTransition(State state, State transition)
		{
			List<State> transitions = state.ChoiceTransitions[state.UISelectedChoice];	
			transitions.Remove(transition);
			UIRefresh();
		}

		private void TriggerSyncAction()
		{
			if (myCurrentState == null || !myCurrentState.WaitForSync)
				return;
			if (myClock <= 0.0f && !myCurrentState.InfiniteDuration)
				Transition();
			else if (myCurrentChoice >= 0)
				Transition();
		}

		private void SwitchStateAction(string newState)
		{
			mySwitchState.valNoCallback = string.Empty;
			State state = GetState(newState);
			if (state != null)
				SwitchState(state);
			else
				SuperController.LogError($"LogicBricks.StateMachine on atom '{containingAtom.uid}': Can't switch to unknown state '{newState}'.");
		}

		private void SetOnCooldownAction(string cooldownState)
		{
			mySetOnCooldown.valNoCallback = string.Empty;
			State state = GetState(cooldownState);
			if (state != null)
				state.Timestamp = Time.time;
			else
				SuperController.LogError($"LogicBricks.StateMachine on atom '{containingAtom.uid}': Can't set cooldown for unknown state '{cooldownState}'.");
		}

		private void SetEnabledAction(string name)
		{
			mySetEnabled.valNoCallback = string.Empty;
			State state = GetState(name);
			if (state != null && !state.Enabled)
			{
				state.Enabled = true;
				myNeedUIRebuild = true;
				myNeedUIRefresh = true;
			}
			else if (state == null)
			{
				SuperController.LogError($"LogicBricks.StateMachine on atom '{containingAtom.uid}': Can't enable unknown state '{name}'.");
			}
		}

		private void SetDisabledAction(string name)
		{
			mySetDisabled.valNoCallback = string.Empty;
			State state = GetState(name);
			if (state != null && state.Enabled)
			{
				state.Enabled = false;
				myNeedUIRebuild = true;
				myNeedUIRefresh = true;
			}
			else if (state == null)
			{
				SuperController.LogError($"LogicBricks.StateMachine on atom '{containingAtom.uid}': Can't disable unknown state '{name}'.");
			}
		}

		private void SetEnabledAllAction(bool allEnabled)
		{
			for (int s=0; s<myStates.Count; ++s)
				myStates[s].Enabled = allEnabled;
			myNeedUIRebuild = true;
			myNeedUIRefresh = true;
		}
		
		private void SetChoiceAction(int choice)
		{
			if (choice >= -CHOICE_NAMES.Count && choice < CHOICE_NAMES.Count)
				myCurrentChoice = choice;
		}

		private void SwitchState(State newState)
		{
			if (myCurrentState != null)
			{
				myCurrentState.Timestamp = Time.time;
				myCurrentState.ExitTrigger.Trigger();
			}

			myCurrentState = newState;
			myCurrentChoice = -1;

			if (myCurrentState != null)
			{
				myCurrentState.EnterTrigger.Trigger();
				if (myCurrentState.InfiniteDuration)
					myClock = myDuration = 1.0f;
				else
					myClock = myDuration = UnityEngine.Random.Range(myCurrentState.MinDuration, myCurrentState.MaxDuration);
				myClockInt = -1;
				myValue = -1.0f;
			}

			UpdateCurrentStateInfo();
			UpdateCurrentStateValue();
		}

		private void Transition()
		{
			int choice = myCurrentChoice >= 0 ? myCurrentChoice : -myCurrentChoice-1;
			List<State> transitions = myCurrentState.ChoiceTransitions[choice];
			
			if (transitions.Count > 0)
			{
				int count = 0;
				for (int i=0; i<transitions.Count; ++i)
				{
					State transition = transitions[i];
					if (transition.Enabled && transition.RemainingCooldown <= 0)
						++count;
				}
				if (count > 0)
				{
					int idx = UnityEngine.Random.Range(0, count);
					for (int i=0, j=0; i<transitions.Count; ++i)
					{
						State transition = transitions[i];
						if (!transition.Enabled || transition.RemainingCooldown > 0)
						{
							continue;
						}
						else if (idx == j)
						{
							SwitchState(transition);
							return;
						}
						else
						{
							++j;
						}
					}
				}

				// fallback to transition whose cooldown has the shortest time
				{
					State minTransition = null;
					float minRemainingCooldown = float.MaxValue;
					for (int i=0; i<transitions.Count; ++i)
					{
						State transition = transitions[i];
						if (!transition.Enabled)
							continue;
						float remainingCooldown = transition.RemainingCooldown;
						if (minRemainingCooldown > remainingCooldown)
						{
							minRemainingCooldown = remainingCooldown;
							minTransition = transition;
						}
					}
					SwitchState(minTransition);
				}
			}
			else
			{
				SwitchState(null);
			}
		}

		private void UpdateCurrentStateInfo()
		{
			if (myCurrentState != null)
			{
				if (myClockInt < 0)
					myCurrentStateInfo.val = "Current State: " + myCurrentState.Name;

				int clockInt = Mathf.CeilToInt(myClock*10.0f);
				if (myCurrentChoice >= 0 && myCurrentState.WaitForSync)
					myCurrentClockInfo.val = "Current Clock: Waiting for TriggerSync";
				else if (myClockInt < 0 && myCurrentState.InfiniteDuration)
					myCurrentClockInfo.val = "Current Clock: Infinite";
				else if (clockInt == 0 && myCurrentState.WaitForSync)
					myCurrentClockInfo.val = "Current Clock: Waiting for TriggerSync";
				else if (myClockInt < 0 || myClockInt != clockInt)
					myCurrentClockInfo.val = string.Format("Current Clock: {0:F1} sec", clockInt/10.0f);
				myClockInt = clockInt;
			}
			else
			{
				myCurrentStateInfo.val = "Current State: ";
				myCurrentClockInfo.val = "Current Clock: ";
			}
		}
		
		private void UpdateCurrentStateValue()
		{
			if (myCurrentState != null)
			{
				float v = 0.0f;
				if (!myCurrentState.InfiniteDuration)
					v = 1.0f - Mathf.Clamp01(myClock / myDuration);
				if (myValue != v)
					myCurrentState.ValueTrigger.Trigger(v);
				myValue = v;
			}
		}

		private void Update()
		{
			for (int i=0; i<myStates.Count; ++i)
			{
				myStates[i].EnterTrigger.Update();
				myStates[i].ValueTrigger.Update();
				myStates[i].ExitTrigger.Update();
			}

			bool uiActive = UITransform.gameObject.activeInHierarchy;
			if (uiActive)
			{
				if (myNeedUIRebuild)
					UIRebuild();
				if (myNeedUIRefresh)
					UIRefresh();
			}

			if (myCurrentState == null || myPaused.val)
				return;
			
			myClock = Mathf.Max(myClock - Time.deltaTime, 0.0f);
			if (myCurrentChoice >= 0 && !myCurrentState.WaitForSync)
			{
				Transition();
			}
			else if (myClock <= 0.0f && !myCurrentState.InfiniteDuration && !myCurrentState.WaitForSync )
			{
				Transition();
			}
			else
			{
				if (uiActive)
					UpdateCurrentStateInfo();
				UpdateCurrentStateValue();
			}
		}

		private void OnAtomRename(string oldid, string newid)
		{
			for (int i=0; i<myStates.Count; ++i)
			{
				myStates[i].EnterTrigger.SyncAtomNames();
				myStates[i].ValueTrigger.SyncAtomNames();
				myStates[i].ExitTrigger.SyncAtomNames();
			}
		}

		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if (includePhysical || forceStore)
			{
				needsStore = true;

				JSONArray sclist = new JSONArray();
				for (int i=0; i<myStates.Count; ++i)
				{
					State state = myStates[i];
					JSONClass sc = new JSONClass();
					sc["Name"] = state.Name;
					sc["Enabled"].AsBool = state.Enabled;
					sc["InfiniteDuration"].AsBool = state.InfiniteDuration;
					sc["WaitForSync"].AsBool = state.WaitForSync;
					sc["MinDuration"].AsFloat = state.MinDuration;
					sc["MaxDuration"].AsFloat = state.MaxDuration;
					sc["Cooldown"].AsFloat = state.Cooldown;
					
					JSONClass ct = new JSONClass();
					for (int c=0; c<CHOICE_NAMES.Count; ++c)
					{
						List<State> transitions = state.ChoiceTransitions[c];
						if (transitions.Count > 0)
						{
							JSONArray tlist = new JSONArray();
							for (int t=0; t<transitions.Count; ++t)
								tlist.Add("", transitions[t].Name);
							ct[CHOICE_NAMES[c]] = tlist;
						}
					}
					sc["ChoiceTransitions"] = ct;
					
					sc[state.EnterTrigger.Name] = state.EnterTrigger.GetJSON(base.subScenePrefix);
					sc[state.ValueTrigger.Name] = state.ValueTrigger.GetJSON(base.subScenePrefix);
					sc[state.ExitTrigger.Name] = state.ExitTrigger.GetJSON(base.subScenePrefix);
					sclist.Add("", sc);
				}
				jc["States"] = sclist;

				if (myCurrentState != null)
				{
					jc["InitialState"] = myCurrentState.Name;
					jc["InitialChoice"].AsInt = myCurrentChoice;
					jc["Clock"].AsFloat = myClock;
					jc["Duration"].AsFloat = myDuration;
				}
				
				jc["ShowAdvOpts"].AsBool = myShowAdvancedOptions.val;
			}

			return jc;
		}

		public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
		{
			base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
			if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
			{
				for (int i=0; i<myStates.Count; ++i)
				{
					myStates[i].EnterTrigger.Remove();
					myStates[i].ValueTrigger.Remove();
					myStates[i].ExitTrigger.Remove();
				}
				myStates.Clear();

				bool hasAdvancedOptions = false;
				JSONArray sclist = jc["States"].AsArray;
				for (int i=0; i<sclist.Count; ++i)
				{
					JSONClass sc = sclist[i].AsObject;
					State state = new State(this, sc["Name"]);
					state.Enabled = !sc.HasKey("Enabled") || sc["Enabled"].AsBool;
					state.InfiniteDuration = sc["InfiniteDuration"].AsBool;
					state.WaitForSync = sc["WaitForSync"].AsBool && !state.InfiniteDuration;
					state.MinDuration = sc["MinDuration"].AsFloat;
					state.MaxDuration = sc["MaxDuration"].AsFloat;
					state.Cooldown = sc["Cooldown"].AsFloat;
					state.EnterTrigger.RestoreFromJSON(sc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
					state.ValueTrigger.RestoreFromJSON(sc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
					state.ExitTrigger.RestoreFromJSON(sc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
					myStates.Add(state);
					
					hasAdvancedOptions |= !state.Enabled || state.WaitForSync || state.Cooldown > 0;
				}

				myStates.Sort(State.SortByNameAscending);
				UIRebuild();

				for (int i=0; i<sclist.Count; ++i)
				{
					JSONClass sc = sclist[i].AsObject;
					State state = GetState(sc["Name"]);
					
					if (sc.HasKey("ChoiceTransitions"))
					{
						JSONClass ct = sc["ChoiceTransitions"].AsObject;
						for (int c=0; c<CHOICE_NAMES.Count; ++c)
						{
							if (!ct.HasKey(CHOICE_NAMES[c]))
								continue;

							List<State> transitions = state.ChoiceTransitions[c];
							JSONArray tlist = ct[CHOICE_NAMES[c]].AsArray;							
							hasAdvancedOptions |= c > 0 && tlist.Count > 0;							
							for (int t=0; t<tlist.Count; ++t)
							{
								State transition = GetState(tlist[t]);
								if (transition != null)
									transitions.Add(transition);
							}
						}
					}
					else // handle legacy
					{						
						List<State> transitions = state.ChoiceTransitions[0];
						JSONArray tlist = sc["Transitions"].AsArray;
						for (int t=0; t<tlist.Count; ++t)
						{
							State transition = GetState(tlist[t]);
							if (transition != null)
								transitions.Add(transition);
						}
					}
				}
				
				myShowAdvancedOptions.valNoCallback = jc.HasKey("ShowAdvOpts") ? jc["ShowAdvOpts"].AsBool : hasAdvancedOptions;

				// switch to initial state, without issuing triggers as these would have been done already before saving.
				myCurrentState = jc.HasKey("InitialState") ? GetState(jc["InitialState"]) : null;
				myCurrentChoice = -1;
				myClock = 0.0f;
				myDuration = 1.0f;
				myClockInt = -1;
				myValue = -1.0f;
				
				if (myCurrentState != null)
				{
					if (jc.HasKey("InitialChoice"))
						myCurrentChoice = Mathf.Clamp(jc["InitialChoice"].AsInt, -1, CHOICE_NAMES.Count-1);
					
					if (myCurrentState.InfiniteDuration)
						myClock = myDuration = 1.0f;
					else
						myClock = myDuration = UnityEngine.Random.Range(myCurrentState.MinDuration, myCurrentState.MaxDuration);
					if (jc.HasKey("Duration"))
						myDuration = jc["Duration"].AsFloat;
					if (jc.HasKey("Clock"))
						myClock = Mathf.Clamp(jc["Clock"].AsFloat, 0.0f, myDuration);
				}
				UpdateCurrentStateInfo();
				UpdateCurrentStateValue();

				List<string> options = myStateChooser.choices;
				if (options.Count == 0)
					myStateChooser.valNoCallback = "";
				else if (myCurrentState != null)
					myStateChooser.valNoCallback = myCurrentState.Name;
				else if (!options.Contains(myStateChooser.val))
					myStateChooser.valNoCallback = options[0];
				myNeedUIRefresh = true;
			}
		}

		private State GetSelectedState()
		{
			return GetState(myStateChooser.val);
		}

		private State GetState(string name)
		{
			int idx = myStateChooser.choices.IndexOf(name);
			if (idx < 0)
				return null;
			else
				return myStates[idx];
		}

		private void OnDestroy()
		{
			SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;

			for (int i=0; i<myStates.Count; ++i)
			{
				myStates[i].EnterTrigger.Remove();
				myStates[i].ValueTrigger.Remove();
				myStates[i].ExitTrigger.Remove();
			}
			myStates.Clear();

			Utils.OnDestroyUI();
		}
    }
}
