﻿using UnityEngine;
using System.Collections.Generic;
using SimpleJSON;

namespace MacGruber
{
    public class ValueThreshold : MVRScript
    {
		private JSONStorableBool myAcceptTrigger;
		private JSONStorableFloat myThresholdValue;
		private JSONStorableStringChooser myThresholdDirection;
		private JSONStorableFloat myValue;
		private EventTrigger myEventTrigger;
		private float myPreviousValue;
		private int myThresholdDirectionIdx = DIRECTION_UP;
		
		private readonly List<string> myThresholdDirectionNames = new List<string>() { "UP", "DOWN", "BOTH" };
		private const int DIRECTION_UP = 0;
		private const int DIRECTION_DOWN = 1;
		private const int DIRECTION_BOTH = 2;
		
        public override void Init()
		{
			Utils.SetupInfoText(this,
				"<color=#606060><size=40><b>ValueThreshold</b></size>\nLogicBrick that triggers when a value crosses a specific threshold.</color>",
				200.0f, true
			);
			
			myValue = Utils.SetupSliderFloatWithRange(this, "Value", 0, 0, 1, false);
			myValue.constrained = false;
			myValue.setCallbackFunction += Trigger;
			myPreviousValue = myValue.val;
			
			myAcceptTrigger = Utils.SetupToggle(this, "AcceptTrigger", true, false);
			
			myThresholdValue = Utils.SetupSliderFloatWithRange(this, "Threshold Value", 0.5f, 0, 1, false);
			myThresholdValue.constrained = false;
						
			myThresholdDirection = Utils.SetupStringChooser(this, "Threshold Direction", myThresholdDirectionNames, false);
			myThresholdDirection.setCallbackFunction += OnSetThresholdDirection;
			OnSetThresholdDirection(myThresholdDirection.val);

			myEventTrigger = new EventTrigger(this, "OnTrigger");
			Utils.SetupButton(this, "Actions OnTrigger", myEventTrigger.OpenPanel, false);

			SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
			SimpleTriggerHandler.LoadAssets();
		}
		
		private void OnSetThresholdDirection(string v)
		{
			myThresholdDirectionIdx = myThresholdDirectionNames.IndexOf(v);
			if (myThresholdDirectionIdx < 0)
				myThresholdDirectionIdx = DIRECTION_UP;
		}

		private void OnAtomRename(string oldid, string newid)
		{
			myEventTrigger.SyncAtomNames();
		}

		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if (includePhysical || forceStore)
			{
				needsStore = true;
				jc[myEventTrigger.Name] = myEventTrigger.GetJSON(base.subScenePrefix);
				jc["PreviousValue"].AsFloat = myPreviousValue;
			}
			return jc;
		}

		public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
		{
			base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
			if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
			{
				myEventTrigger.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
			}
			
			myPreviousValue = jc.HasKey("PreviousValue") ? jc["PreviousValue"].AsFloat : myValue.val;
		}

		private void OnDestroy()
		{
			SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;
			myEventTrigger.Remove();
		}
		
		private void Update()
		{
			myEventTrigger.Update();
		}

		private void Trigger(float v)
		{			
			if (myAcceptTrigger.val)
			{
				float threshold = myThresholdValue.val;
				if (myThresholdDirectionIdx != DIRECTION_DOWN && myPreviousValue < threshold && v >= threshold)
					myEventTrigger.Trigger();
				else if (myThresholdDirectionIdx != DIRECTION_UP && myPreviousValue > threshold && v <= threshold)
					myEventTrigger.Trigger();
			}
			myPreviousValue = v;
		}
    }
}
