﻿using UnityEngine;
using System.Collections.Generic;
using SimpleJSON;

namespace MacGruber
{
    public class ValueUpDown : MVRScript
    {
		private JSONStorableFloat myValue;
		private JSONStorableBool mySmoothStep;
		private JSONStorableFloat mySpeed;
		private JSONStorableFloat mySpeedAdjust;
		private JSONStorableFloat myPauseSpeed;
		private FloatTrigger myFloatTrigger;
		private EventTrigger myPauseTrigger;		

		private float myStartDistance = -1.0f;
		private float myEndDistance = 0.0f;
		private float myTraveledDistance = 0.0f;
		private float mySpeedStart = 0.0f;
		private float mySpeedEnd = 0.0f;
		private float mySpeedActual = 0.0f;
		private bool myDirection = true;
		private bool myPausedDesired = true;

        public override void Init()
		{
			Utils.SetupInfoText(this,
				"<color=#606060><size=40><b>ValueUpDown</b></size>\nLogicBrick that provides a value continuously moving up and down between 0 and 1. "+
				"Via triggers it can be paused precisely in up or down position.</color>",
				300.0f, true
			);

			Utils.SetupAction(this, "PauseUp", PauseUp);
			Utils.SetupAction(this, "PauseDown", PauseDown);
			Utils.SetupAction(this, "Unpause", Unpause);

			myValue = Utils.SetupSliderFloat(this, "Value", 0, 0, 1, false);
			myValue.setCallbackFunction += Trigger;
			mySmoothStep = Utils.SetupToggle(this, "Use SmoothStep Curve", true, false);			
			mySpeed = Utils.SetupSliderFloat(this, "Speed", 1, 0, 10, false);
			mySpeed.setCallbackFunction += Speed;
			mySpeedAdjust = Utils.SetupSliderFloat(this, "Speed Adjust Time", 1, 0, 10, false);
			myPauseSpeed = Utils.SetupSliderFloat(this, "Start/Stop Speed", 0.3f, 0, 10, false);

			myFloatTrigger = new FloatTrigger(this, "OnValueSet");
			Utils.SetupButton(this, "Actions OnValueSet", myFloatTrigger.OpenPanel, false);
			
			myPauseTrigger = new EventTrigger(this, "OnPauseTrigger");
			Utils.SetupButton(this, "Actions OnPauseTrigger", myPauseTrigger.OpenPanel, false);

			Utils.SetupButton(this, "Test PauseUp", PauseUp, false);
			Utils.SetupButton(this, "Test PauseDown", PauseDown, false);
			Utils.SetupButton(this, "Test Unpause", Unpause, false);

			SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
			SimpleTriggerHandler.LoadAssets();
		}

		private void OnAtomRename(string oldid, string newid)
		{
			myFloatTrigger.SyncAtomNames();
			myPauseTrigger.SyncAtomNames();
		}

		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if (includePhysical || forceStore)
			{
				needsStore = true;
				jc["SpeedActual"].AsFloat = mySpeedActual;
				jc["Direction"].AsBool = myDirection;
				jc[myFloatTrigger.Name] = myFloatTrigger.GetJSON(base.subScenePrefix);
				jc[myPauseTrigger.Name] = myPauseTrigger.GetJSON(base.subScenePrefix);
			}
			return jc;
		}

		public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
		{
			base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
			if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
			{
				mySpeedActual = jc["SpeedActual"].AsFloat;
				myDirection = jc["Direction"].AsBool;
				myPausedDesired = mySpeedActual <= 0.0f;
				if (!myPausedDesired)
					Speed(mySpeed.val);

				myFloatTrigger.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
				myPauseTrigger.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
			}
		}

		private void OnDestroy()
		{
			SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;
			myFloatTrigger.Remove();
			myPauseTrigger.Remove();
		}

		private void Update()
		{
			myFloatTrigger.Update();
			myPauseTrigger.Update();

			if (mySpeedActual <= 0.0f)
				return;

			float t = mySpeedActual * Time.deltaTime;
			myTraveledDistance = Mathf.Min(myTraveledDistance + t, myEndDistance);
			float s = Mathf.InverseLerp(myStartDistance, myEndDistance, myTraveledDistance);
			mySpeedActual = Mathf.Lerp(mySpeedStart, mySpeedEnd, s);
			float v = myValue.val + t * (myDirection ? 1 : -1);

			if ((myDirection && v >= 1.0f) || (!myDirection && v <= 0.0f))
			{
				if (myPausedDesired && myEndDistance - myTraveledDistance < 0.5f)
				{
					v = myDirection ? 1.0f : 0.0f;
					mySpeedActual = 0.0f;
					myPauseTrigger.Trigger();
				}
				else
				{
					v = myDirection ? Mathf.Max(2.0f - v, 0.0f) : Mathf.Min(-v, 1.0f);
				}
				myDirection = !myDirection;
			}

			myValue.val = v;
		}

		private void PauseUp()
		{
			if (!myPausedDesired)
				PauseAt(true);
		}

		private void PauseDown()
		{
			if (!myPausedDesired)
				PauseAt(false);
		}

		private void PauseAt(bool pauseDirection)
		{
			mySpeedStart = mySpeedActual;
			mySpeedEnd = Mathf.Min(myPauseSpeed.val, mySpeedActual);
			float distance = (mySpeedStart + mySpeedEnd) * 0.5f * mySpeedAdjust.val;
			int i = Mathf.FloorToInt(distance);
			float vs = (i % 2) == 0 ? 1.0f - (distance - i) : (i - distance);
			float vc = pauseDirection ? myValue.val : 1.0f - myValue.val;
			vc *= (pauseDirection ^ myDirection) ? -1 : 1;

			myStartDistance = vs - vc;
			if (vs <= vc)
				myStartDistance += 2.0f;
			myEndDistance = myStartDistance + distance;
			myTraveledDistance = 0.0f;
			myPausedDesired = true;
		}

		private void Unpause()
		{
			if (!myPausedDesired)
				return;

			mySpeedActual = Mathf.Max(myPauseSpeed.val, mySpeedActual);
			mySpeedStart = mySpeedActual;
			mySpeedEnd = mySpeed.val;
			float distance = (mySpeedStart + mySpeedEnd) * 0.5f * mySpeedAdjust.val;

			myStartDistance = 0.0f;
			myEndDistance = distance;
			myTraveledDistance = 0.0f;
			myPausedDesired = false;
		}

		private void Speed(float v)
		{
			if (myPausedDesired)
				return;

			mySpeedStart = mySpeedActual;
			mySpeedEnd = v;
			float distance = (mySpeedStart + mySpeedEnd) * 0.5f * mySpeedAdjust.val;

			myStartDistance = 0.0f;
			myEndDistance = distance;
			myTraveledDistance = 0.0f;
		}

		private void Trigger(float v)
		{
			if (mySmoothStep.val)
				v = Mathf.SmoothStep(0, 1, v);
			myFloatTrigger.Trigger(v);
		}
    }
}
